# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import FrozenInstanceError
from datetime import datetime
from unittest import TestCase

from notus.scanner.models.advisory import (
    Advisory,
    Severity,
)


class AdvisoryTestCase(TestCase):
    def test_defaults(self):
        severity_date = datetime(year=2021, month=1, day=2, hour=11, minute=11)
        creation_date = datetime(year=2021, month=3, day=21, hour=10, minute=0)
        last_modification = datetime.now()
        cvss_v3 = "CVSS:3.1/AV:N/AC:L/PR:L/UI:N/S:U/C:L/I:L/A:L"
        severity = Severity(
            origin="foo",
            date=severity_date,
            cvss_v3=cvss_v3,
        )
        advisory = Advisory(
            oid="1.2.3.4.5",
            title="Foo Bar",
            creation_date=creation_date,
            last_modification=last_modification,
            advisory_id="1234",
            advisory_xref="http://foo/1234",
            severity=severity,
        )

        self.assertEqual(advisory.oid, "1.2.3.4.5")
        self.assertEqual(advisory.title, "Foo Bar")
        self.assertEqual(advisory.creation_date, creation_date)
        self.assertEqual(advisory.last_modification, last_modification)
        self.assertEqual(advisory.advisory_id, "1234")
        self.assertEqual(advisory.advisory_xref, "http://foo/1234")
        self.assertEqual(advisory.severity.origin, "foo")
        self.assertEqual(advisory.severity.date, severity_date)
        self.assertIsNone(advisory.severity.cvss_v2)
        self.assertEqual(advisory.severity.cvss_v3, cvss_v3)
        self.assertIsNone(advisory.summary)
        self.assertIsNone(advisory.insight)
        self.assertIsNone(advisory.affected)
        self.assertIsNone(advisory.impact)

        self.assertIsInstance(advisory.cves, list)
        self.assertEqual(len(advisory.cves), 0)
        self.assertIsInstance(advisory.xrefs, list)
        self.assertEqual(len(advisory.xrefs), 0)

    def test_hash(self):
        severity_date = datetime(year=2021, month=1, day=2, hour=11, minute=11)
        creation_date = datetime(year=2021, month=3, day=21, hour=10, minute=0)
        last_modification = datetime.now()
        cvss_v3 = "CVSS:3.1/AV:N/AC:L/PR:L/UI:N/S:U/C:L/I:L/A:L"
        severity = Severity(
            origin="foo",
            date=severity_date,
            cvss_v3=cvss_v3,
        )
        advisory1 = Advisory(
            oid="1.2.3.4.5",
            title="Foo Bar",
            creation_date=creation_date,
            last_modification=last_modification,
            advisory_id="1234",
            advisory_xref="http://foo/1234",
            severity=severity,
        )
        advisory2 = Advisory(
            oid="1.2.3.4.5",
            title="Bar Foo",
            creation_date=creation_date,
            last_modification=last_modification,
            advisory_id="4321",
            advisory_xref="http://bar/4321",
            severity=severity,
        )
        advisory3 = Advisory(
            oid="1.2.3.4.6",
            title="Foo Bar",
            creation_date=creation_date,
            last_modification=last_modification,
            advisory_id="1234",
            advisory_xref="http://foo/1234",
            severity=severity,
        )

        self.assertEqual(hash(advisory1), hash(advisory2))
        self.assertNotEqual(hash(advisory1), hash(advisory3))

    def test_immutability(self):
        severity_date = datetime(year=2021, month=1, day=2, hour=11, minute=11)
        creation_date = datetime(year=2021, month=3, day=21, hour=10, minute=0)
        last_modification = datetime.now()
        cvss_v3 = "CVSS:3.1/AV:N/AC:L/PR:L/UI:N/S:U/C:L/I:L/A:L"
        severity = Severity(
            origin="foo",
            date=severity_date,
            cvss_v3=cvss_v3,
        )
        advisory = Advisory(
            oid="1.2.3.4.5",
            title="Foo Bar",
            creation_date=creation_date,
            last_modification=last_modification,
            advisory_id="1234",
            advisory_xref="http://foo/1234",
            severity=severity,
        )

        with self.assertRaises(FrozenInstanceError):
            advisory.oid = "2.2.2.2.2"

        with self.assertRaises(FrozenInstanceError):
            advisory.title = "Ipsum"

        with self.assertRaises(FrozenInstanceError):
            advisory.last_modification = datetime.now()



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod
import logging
from enum import Enum
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Set, Optional

from ...errors import PackageError


logger = logging.getLogger(__name__)


class PackageType(Enum):
    RPM = "rpm"
    DEB = "deb"
    EBUILD = "ebuild"


class PackageComparision(Enum):
    EQUAL = 0  # a and b are equal
    A_NEWER = 1  # a is newer than b
    B_NEWER = 2  # b is newer than a
    # a and b are not compareable. e.g. different architectures
    NOT_COMPARABLE = 3


class Architecture(Enum):
    NOARCH = "noarch"
    SOURCE = "src"
    ALL = "all"
    I386 = "i386"
    I586 = "i586"
    I686 = "i686"
    X86_64 = "x86_64"
    ARMV7L = "armv7l"
    ARMV6L = "armv6l"
    AARCH64 = "aarch64"
    AMD64 = "amd64"
    IA64 = "ia64"
    PPC = "ppc"
    PPC64 = "ppc64"
    PPC64LE = "ppc64le"
    S390 = "s390"
    S390X = "s390x"
    AARCH64_ILP32 = "aarch64_ilp32"
    UNKNOWN = "UNKNOWN"
    NOTSET = "NOTSET"


@dataclass
class Package:
    "Base class for different Package types"
    name: str
    full_name: str
    full_version: str

    def __guard_cmp(self, other: Any):
        if not isinstance(other, type(self)):
            raise PackageError(f"Can't compare {self!r} to {other!r}.")

    def __gt__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self._compare(other) == PackageComparision.A_NEWER

    def __lt__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self._compare(other) == PackageComparision.B_NEWER

    def __eq__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self._compare(other) == PackageComparision.EQUAL

    def __ge__(self, other: Any) -> bool:
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other: Any) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __hash__(self) -> int:
        # allow to hash the package
        # the full name identifies the package
        return hash(self.full_name)

    @abstractmethod
    def _compare(self, other: Any) -> PackageComparision:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_full_name(full_name: str):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_name_and_full_version(name: str, full_version: str):
        raise NotImplementedError()


@dataclass(frozen=True)
class Verifier:
    symbol: str
    _verifier: Callable[[Package, Package], bool]

    def __name__(self) -> str:
        return self.symbol

    def verify(self, expected: Package, actual: Package) -> bool:
        return self._verifier(expected, actual)


@dataclass(frozen=True)
class AdvisoryReference:
    """A reference to a vulnerability advisory"""

    oid: str


@dataclass(frozen=True, unsafe_hash=True)
class PackageAdvisory:
    """Connects a package with an advisory"""

    package: Package
    advisory: AdvisoryReference
    is_vulnerable: Callable[[Package], bool] = field(compare=False, hash=False)


@dataclass(frozen=True)
class PackageAdvisories:
    """Container for mapping a package name to a set of advisories for this
    package"""

    package_type: PackageType
    advisories: Dict[str, Set[PackageAdvisory]] = field(default_factory=dict)

    def get_package_advisories_for_package(
        self, package: Package
    ) -> Set[PackageAdvisory]:
        return self.advisories.get(package.name) or set()

    @staticmethod
    def is_vulnerable_from_symbol(symbol: Optional[str]):
        """
        is_vulnerable_from_symbol returns either a Verifier when the symbol
        identifier contains a known operand or or >= if not.
        """
        if not symbol or symbol.startswith(">="):
            return Verifier(">=", lambda a, b: a > b)
        if symbol.startswith("<="):
            return Verifier("<=", lambda a, b: a < b)
        if symbol.startswith("="):
            return Verifier("=", lambda a, b: a != b)
        if symbol.startswith("<"):
            return Verifier("<", lambda a, b: a <= b)
        if symbol.startswith(">"):
            return Verifier(">", lambda a, b: a >= b)

        return Verifier(">=", lambda a, b: a > b)

    def add_advisory_for_package(
        self,
        package: Package,
        advisory: AdvisoryReference,
        verifier: Optional[str],
    ) -> None:

        advisories = self.get_package_advisories_for_package(package)
        use_verifier = self.is_vulnerable_from_symbol(verifier)
        is_vulnerable = lambda other: use_verifier.verify(package, other)

        advisories.add(PackageAdvisory(package, advisory, is_vulnerable))
        self.advisories[package.name] = advisories

    def __len__(self) -> int:
        return len(self.advisories)



# slightly adjusted from https://github.com/ihiji/version_utils
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import re
from dataclasses import dataclass
from packaging.version import parse

from .package import (
    Package,
    Architecture,
    PackageComparision,
)

_rpm_re = re.compile(r"(\S+)-(?:(\d*):)?(.*)-(~?\w+[\w.]*)")

_rpm_compile_no_arch = re.compile("(.*)-([^-]+)-([^-]+)")
_rpm_compile = re.compile(r"(.*)-([^-]+)-([^-]+)\.([^-]+)")
_rpm_compile_version = re.compile(r"([^-]+)-([^-]+)\.([^-]+)")

logger = logging.getLogger(__name__)


@dataclass
class RPMPackage(Package):
    """Represents a RPM package"""

    version: str
    release: str
    arch: str

    __hash__ = Package.__hash__

    def _compare(self, other: "RPMPackage") -> PackageComparision:
        if self.arch != other.arch:
            return PackageComparision.NOT_COMPARABLE

        if self.full_version == other.full_version:
            return PackageComparision.EQUAL

        a_ver = parse(self.version)
        b_ver = parse(other.version)
        if a_ver != b_ver:
            return (
                PackageComparision.A_NEWER
                if a_ver > b_ver
                else PackageComparision.B_NEWER
            )
        a_rel = parse(self.release)
        b_rel = parse(other.release)
        return (
            PackageComparision.A_NEWER
            if a_rel > b_rel
            else PackageComparision.B_NEWER
        )

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None

        full_name = full_name.strip()

        try:
            name, version, release, architecture = _rpm_compile.match(
                full_name
            ).groups()
            try:
                arch = Architecture(architecture)
            except ValueError:
                arch = Architecture.UNKNOWN
        except AttributeError:
            try:
                name, version, release = _rpm_compile_no_arch.match(
                    full_name
                ).groups()
                arch = Architecture.NOTSET
            except AttributeError:
                logger.warning(
                    "The rpm package %s could not be parsed", full_name
                )
                return None

        return RPMPackage(
            name=name,
            version=version,
            release=release,
            arch=arch,
            full_name=full_name,
            full_version=f"{version}-{release}.{arch.value}",
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None

        name = name.strip()
        full_version = full_version.strip()

        version_match = _rpm_compile_version.match(full_version)
        if not version_match:
            return None
        version, release, architecture = version_match.groups()

        try:
            arch = Architecture(architecture)
        except ValueError:
            arch = Architecture.UNKNOWN

        return RPMPackage(
            name=name,
            version=version,
            release=release,
            arch=arch,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
        )



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Dict

from ..models.advisory import Advisory
from .packages.package import Package


@dataclass
class PackageVulnerability:
    """Represents a vulnerable package on a host"""

    host_ip: str
    host_name: str
    packages: Dict[Package, Package]
    advisory: Advisory

    def __init__(
        self,
        host_ip: str,
        host_name: str,
        package: Package,
        fixed_package: Package,
        advisory: Advisory,
    ) -> None:
        self.host_ip = host_ip
        self.host_name = host_name
        self.packages = {}
        self.packages[package] = fixed_package
        self.advisory = advisory

    def add_package(self, package: Package, fixed_package: Package):
        self.packages[package] = fixed_package



# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
from typing import List, Optional, Tuple
import unittest
from pathlib import Path


from notus.scanner.messages.start import ScanStartMessage
from notus.scanner.messages.message import Message
from notus.scanner.messaging.publisher import Publisher
from notus.scanner.scanner import NotusScanner
from notus.scanner.loader.gpg_sha_verifier import VerificationResult
from notus.scanner.loader.json import JSONAdvisoriesLoader

_here = Path(__file__).parent


class FakePublisher(Publisher):
    def __init__(self):
        self.results = []

    def publish(self, message: Message) -> None:
        serialized = message.serialize()
        values = str(serialized.get("value", "")).split("\n")
        for value in values:
            if value.find("Installed version:") >= 0:
                installed = value[len("Installed version: ") :].strip()
                self.results.append(installed)


class VerifierTestCase(unittest.TestCase):
    """
    VerifierTestCase loads a notus advisory for FakeSpecifierOS
    and iterates through defined fixed packages and their specifier
    to generate and test them.
    """

    def per_symbol(
        self, name: str, verifier: Optional[str]
    ) -> Tuple[List[str], List[str]]:
        """
        returns not_in_result and in result
        """
        greater = lambda: name.replace("15", "16")
        smaller = lambda: name.replace("15", "14")
        if not verifier:
            return [greater(), name], [smaller()]
        if verifier == ">":
            return [greater()], [name, smaller()]
        if verifier == ">=":
            return [greater(), name], [smaller()]
        if verifier == "<":
            return [smaller()], [name, greater()]
        if verifier == "<=":
            return [smaller(), name], [greater()]
        return [name], [smaller(), greater()]

    def generate_test_cases(self):
        jdict = json.loads((_here / "fakespecifier_os.notus").read_bytes())
        fixed_packages = (
            f.get("fixed_packages", []) for f in jdict.get("advisories", {})
        )
        cases = (
            (fp.get("full_name"), fp.get("specifier"))
            for fps in fixed_packages
            for fp in fps
        )
        # cases that should not appear in result
        not_in = []
        # cases that should appear in result
        is_in = []
        for verylongvariablenamecaseforpylint in cases:
            c_not_in, c_is_in = self.per_symbol(
                *verylongvariablenamecaseforpylint
            )
            not_in = not_in + c_not_in
            is_in = is_in + c_is_in
        # packagelist is both combined
        return not_in + is_in, not_in, is_in

    def test_verifier(self):
        loader = JSONAdvisoriesLoader(
            advisories_directory_path=_here,
            verify=lambda _: VerificationResult.SUCCESS,
        )
        publisher = FakePublisher()
        scanner = NotusScanner(loader, publisher)
        pkg_list, not_in_results, in_results = self.generate_test_cases()
        msg = ScanStartMessage(
            scan_id="scanus praktikus",
            os_release="FakeSpecifier OS",
            host_ip="127.0.0.2",
            host_name="localhorst",
            package_list=pkg_list,
        )
        scanner.run_scan(msg)
        results = set(publisher.results)
        self.assertEqual(len(results), len(publisher.results))

        self.assertEqual(set(), results.intersection(not_in_results))

        self.assertEqual(set(in_results), results.intersection(in_results))



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from unittest import TestCase

from notus.scanner.models.advisory import Advisory, Severity
from notus.scanner.models.packages.rpm import RPMPackage
from notus.scanner.models.vulnerability import PackageVulnerability


class PackageVulnerabilityTestCase(TestCase):
    def test_constructor(self):
        host_ip = "123.456.789.10"
        host_name = "foo"

        severity_date = datetime(year=2021, month=1, day=2, hour=11, minute=11)
        creation_date = datetime(year=2021, month=3, day=21, hour=10, minute=0)
        last_modification = datetime.now()
        severity = Severity(
            origin="foo",
            date=severity_date,
        )
        advisory = Advisory(
            oid="1.2.3.4.5",
            title="Foo Bar",
            creation_date=creation_date,
            last_modification=last_modification,
            advisory_id="1234",
            advisory_xref="http://foo/1234",
            severity=severity,
        )

        package = RPMPackage.from_full_name(
            "foo-bar-1.2.3-4.x86_64",
        )

        fixed_package = RPMPackage.from_full_name(
            "foo-bar-1.2.4-4.x86_64",
        )

        package_vulnerability = PackageVulnerability(
            host_ip=host_ip,
            host_name=host_name,
            advisory=advisory,
            package=package,
            fixed_package=fixed_package,
        )

        self.assertEqual(package_vulnerability.host_ip, host_ip)
        self.assertEqual(package_vulnerability.host_name, host_name)

        advisory = package_vulnerability.advisory
        self.assertEqual(advisory.oid, "1.2.3.4.5")

        package, fixed_package = next(
            iter(package_vulnerability.packages.items())
        )
        self.assertEqual(package.name, "foo-bar")
        self.assertEqual(package.version, "1.2.3")  # pylint: disable=no-member
        self.assertEqual(package.release, "4")  # pylint: disable=no-member
        self.assertEqual(package.full_name, "foo-bar-1.2.3-4.x86_64")

        self.assertEqual(fixed_package.name, "foo-bar")
        self.assertEqual(
            fixed_package.version, "1.2.4"  # pylint: disable=no-member
        )
        self.assertEqual(
            fixed_package.release, "4"  # pylint: disable=no-member
        )
        self.assertEqual(fixed_package.full_name, "foo-bar-1.2.4-4.x86_64")



# Copyright (C) 2014-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys

from pathlib import Path
from typing import Callable, Dict, Optional

from .cli import CliParser
from .errors import AdvisoriesLoadingError, Sha256SumLoadingError
from .loader import JSONAdvisoriesLoader
from .messaging.mqtt import (
    MQTTDaemon,
    MQTTPublisher,
    MQTTClient,
    MQTTSubscriber,
)
from .messages.start import ScanStartMessage
from .scanner import NotusScanner
from .utils import (
    go_to_background,
    create_pid,
    init_signal_handler,
    init_logging,
)

from .loader.gpg_sha_verifier import (
    ReloadConfiguration,
    VerificationResult,
    create_verify,
    reload_sha256sums,
)

from .__version__ import __version__


logger = logging.getLogger(__name__)

SENTRY_DSN_NOTUS_SCANNER = os.environ.get("SENTRY_DSN_NOTUS_SCANNER")
if SENTRY_DSN_NOTUS_SCANNER:
    import sentry_sdk  # pylint: disable=import-error

    sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
        SENTRY_DSN_NOTUS_SCANNER,
        traces_sample_rate=1.0,
        server_name=os.environ.get("SENTRY_SERVER_NAME"),
        environment=os.environ.get("SENTRY_ENVIRONMENT"),
    )


def hashsum_verificator(
    products_directory_path: Path, disable: bool
) -> Callable[[Path], VerificationResult]:
    if disable:
        logger.info("hashsum verification is disabled")
        return lambda _: VerificationResult.SUCCESS

    def on_hash_sum_verification_failure(
        _: Optional[Dict[str, str]]
    ) -> Dict[str, str]:
        raise Sha256SumLoadingError(
            f"Unable to verify signature of {sha_sum_file_path}"
        )

    sha_sum_file_path = products_directory_path / "sha256sums"
    sha_sum_reload_config = ReloadConfiguration(
        hash_file=sha_sum_file_path,
        on_verification_failure=on_hash_sum_verification_failure,
    )

    sums = reload_sha256sums(sha_sum_reload_config)
    return create_verify(sums)


def run_daemon(
    mqtt_broker_address: str,
    mqtt_broker_port: int,
    products_directory_path: Path,
    disable_hashsum_verification: bool,
):
    """Initialize the mqtt client, mqtt handler, notus scanner and run
    forever
    """

    if not products_directory_path.is_dir():
        raise AdvisoriesLoadingError(
            f"Can't load advisories. {products_directory_path.absolute()} is"
            " not a directory."
        )

    verifier = hashsum_verificator(
        products_directory_path, disable_hashsum_verification
    )

    loader = JSONAdvisoriesLoader(
        advisories_directory_path=products_directory_path, verify=verifier
    )
    try:
        client = MQTTClient(
            mqtt_broker_address=mqtt_broker_address,
            mqtt_broker_port=mqtt_broker_port,
        )
    except ConnectionRefusedError:
        logger.error(
            "Could not connect to MQTT broker at %s. Connection refused.",
            mqtt_broker_address,
        )
        sys.exit(1)

    daemon = MQTTDaemon(client)

    publisher = MQTTPublisher(client)
    scanner = NotusScanner(loader=loader, publisher=publisher)

    subscriber = MQTTSubscriber(client)
    subscriber.subscribe(ScanStartMessage, scanner.run_scan)

    daemon.run()


def main():
    parser = CliParser()
    args = parser.parse_arguments()

    init_logging(
        "notus-scanner",
        args.log_level,
        log_file=args.log_file,
        foreground=args.foreground,
    )

    if not args.foreground:
        go_to_background()

    if not create_pid(args.pid_file):
        sys.exit()

    init_signal_handler(args.pid_file)

    logger.info("Starting notus-scanner version %s.", __version__)

    run_daemon(
        args.mqtt_broker_address,
        args.mqtt_broker_port,
        args.products_directory,
        args.disable_hashsum_verification,
    )


if __name__ == "__main__":
    main()



# slightly adjusted from https://github.com/ihiji/version_utils
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
deb module for version_utils

Contains dpkg parsing and comparison operations for version_utils.
Public methods include:

    * :any:`compare_packages`: compare two dpkg package strings, e.g.
      ``gcc-4.4.7-16.el6.x86_64`` and ``gcc-4.4.7-17.el6.x86_64``
    * :any:`compare_versions`: compare two dpkg version strings (the
      bit between the dashes in an dpkg package string)
    * :any:`package`: parse an dpkg package string to get name, epoch,
      version, release, and architecture information. Returns as a
      :any:`common.Package` object.
"""

import logging
import re
from typing import Tuple
from dataclasses import dataclass
from packaging.version import parse

from .package import Package, PackageComparision

_deb_compile = re.compile(r"(.*)-(?:(\d*):)?(.*)-(.*)")
_deb_compile_wo_revision = re.compile(r"(.*)-(?:(\d*):)?(.*)")
_deb_compile_version = re.compile(r"(?:(\d*):)?(\d.*)-(.*)")
_deb_compile_version_wo_revision = re.compile(r"(?:(\d*):)?(\d.*)")


logger = logging.getLogger(__name__)

# Epoch, Version, Revision Tuple
EVRTuple = Tuple[str, str, str]


@dataclass
class DEBPackage(Package):
    """Represents a .deb based package"""

    upstream_version: str
    debian_revision: str
    epoch: str

    __hash__ = Package.__hash__

    def _compare(self, other: Package) -> PackageComparision:
        a_version = parse(self.full_version)
        b_version = parse(other.full_version)

        if a_version == b_version:
            return PackageComparision.EQUAL

        return (
            PackageComparision.A_NEWER
            if a_version > b_version
            else PackageComparision.B_NEWER
        )

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None

        full_name = full_name.strip()
        # Try to get data with
        try:
            name, epoch, upstream_version, debian_revision = _deb_compile.match(
                full_name
            ).groups()
        except AttributeError:
            try:
                name, epoch, upstream_version = _deb_compile_wo_revision.match(
                    full_name
                ).groups()
                debian_revision = ""
            except AttributeError:
                logger.warning(
                    "The deb package %s could not be parsed", full_name
                )
                return None

        if not epoch:
            epoch = "0"
            full_version = f"{upstream_version}"
        else:
            full_version = f"{epoch}:{upstream_version}"

        if debian_revision:
            full_version = f"{full_version}-{debian_revision}"

        return DEBPackage(
            name=name,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
            epoch=epoch,
            upstream_version=upstream_version,
            debian_revision=debian_revision,
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None
        name = name.strip()
        full_version = full_version.strip()
        try:
            (
                epoch,
                upstream_version,
                debian_revision,
            ) = _deb_compile_version.match(full_version).groups()
        except AttributeError:
            try:
                (
                    epoch,
                    upstream_version,
                ) = _deb_compile_version_wo_revision.match(
                    full_version
                ).groups()
                debian_revision = ""
            except AttributeError:
                logger.warning(
                    "The deb package %s %s could not be parsed",
                    name,
                    full_version,
                )
                return None

        if not epoch:
            epoch = "0"

        return DEBPackage(
            name=name,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
            epoch=epoch,
            upstream_version=upstream_version,
            debian_revision=debian_revision,
        )



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional


@dataclass(frozen=True)
class Severity:
    origin: str
    date: datetime
    cvss_v2: Optional[str] = None
    cvss_v3: Optional[str] = None


@dataclass(frozen=True)
class Advisory:
    oid: str
    title: str
    creation_date: datetime
    last_modification: datetime
    advisory_id: str
    advisory_xref: str
    severity: Severity
    summary: Optional[str] = None
    insight: Optional[str] = None
    affected: Optional[str] = None
    impact: Optional[str] = None
    cves: List[str] = field(default_factory=list)
    xrefs: List[str] = field(default_factory=list)

    def __hash__(self) -> int:
        # implement hash function for storage in dict, set, etc.
        # use oid for hashing because it is the unique id of an advisory
        # different advisories must have a different oid
        return hash(self.oid)



"""
ebuild module for version_utils

Currently the ebuild are based upon find within `/usr/portage` and are
manipulated as:
sed 's,/usr/portage/([^/]*)/.*/([^/]*).ebuild$,\1/\2,'

Meaning the corresponding name will be:
x11-plugins/wmix-3.5

The version generated by our generater will be delivered via name and version.

The version of the program is program specific hence we could just use Version
of pep440 to guess immediately.
"""

import logging
from dataclasses import dataclass
from packaging.version import Version

from .package import Package, PackageComparision


logger = logging.getLogger(__name__)


@dataclass(eq=False)
class EBuildPackage(Package):
    """Represents a .ebuild based package"""

    def _compare(self, other: Package) -> PackageComparision:
        # we don't want to deal with LegacyVersion
        a_version = Version(self.full_version)
        b_version = Version(other.full_version)

        if self.name != other.name:
            return PackageComparision.NOT_COMPARABLE

        if a_version == b_version:
            return PackageComparision.EQUAL

        return (
            PackageComparision.A_NEWER
            if a_version > b_version
            else PackageComparision.B_NEWER
        )

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None
        base_name = (
            full_name[full_name.index("/") + 1 :]
            if "/" in full_name
            else full_name
        )
        if not base_name:
            return None
        d_index = base_name.index("-") if "-" in base_name else len(base_name)
        full_version = base_name[d_index + 1 :]
        if not full_version:
            return None
        return EBuildPackage(
            name=full_name[: len(full_name) - len(base_name[d_index:])],
            full_name=full_name,
            full_version=full_version,
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None
        name = name.strip()
        full_version = full_version.strip()
        return EBuildPackage(
            name=name,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['notus',
 'notus.scanner',
 'notus.scanner.cli',
 'notus.scanner.loader',
 'notus.scanner.messages',
 'notus.scanner.messaging',
 'notus.scanner.models',
 'notus.scanner.models.packages',
 'notus.scanner.tools',
 'tests',
 'tests.cli',
 'tests.loader',
 'tests.messages',
 'tests.messaging',
 'tests.models',
 'tests.models.packages']

package_data = \
{'': ['*']}

modules = \
['poetry']
install_requires = \
['packaging<21.3',
 'paho-mqtt>=1.6,<2.0',
 'psutil>=5.9,<6.0',
 'python-gnupg>=0.4.8,<0.5.0',
 'tomli<3.0.0']

extras_require = \
{'sentry': ['sentry-sdk>=1.6.0,<2.0.0']}

entry_points = \
{'console_scripts': ['notus-scan-start = notus.scanner.tools.scanstart:main',
                     'notus-scanner = notus.scanner.daemon:main',
                     'notus-subscriber = notus.scanner.tools.subscriber:main']}

setup_kwargs = {
    'name': 'notus-scanner',
    'version': '22.4.0',
    'description': 'A vulnerability scanner for creating results from local security checks (LSCs) ',
    'long_description': '![Greenbone Logo](https://www.greenbone.net/wp-content/uploads/gb_new-logo_horizontal_rgb_small.png)\n\n# Notus Scanner <!-- omit in toc -->\n\n[![Build and test](https://github.com/greenbone/notus-scanner/actions/workflows/ci-python.yml/badge.svg)](https://github.com/greenbone/notus-scanner/actions/workflows/ci-python.yml)\n[![codecov](https://codecov.io/gh/greenbone/notus-scanner/branch/main/graph/badge.svg?token=LaduLacbWO)](https://codecov.io/gh/greenbone/notus-scanner)\n\nNotus Scanner detects vulnerable products in a system environment. The scanning\nmethod is to evaluate internal system information. It does this very fast and\neven detects currently inactive products because it does not need to interact\nwith each of the products.\n\nTo report about vulnerabilities, Notus Scanner receives collected system\ninformation on the one hand and accesses the vulnerability information from the\nfeed service on the other. Both input elements are in table form: the system\ninformation is specific to each environment and the vulnerability information is\nspecific to each system type.\n\nNotus Scanner integrates into the Greenbone Vulnerability Management framework\nwhich allows to let it scan entire networks within a single task. Any\nvulnerability test in the format of `.notus` files inside the Greenbone Feed\nwill be considered and automatically matched with the scanned environments.\n\nA system environment can be the operating system of a host. But it could also be\ncontainers like Docker or virtual machines. Neither of these need to be actively\nrunning for scanning.\n\nThe Notus Scanner is implemented in Python and published under an Open Source\nlicense. Greenbone Networks maintains and extends it since it is embedded in the\nGreenbone Professional Edition as well as in the Greenbone Cloud Services.\n\nGreenbone also keeps the vulnerability information up-to-date via the feed on a\ndaily basis. The `.notus` format specification is open and part of the\ndocumentation.\n\n## Table of Contents <!-- omit in toc -->\n\n- [Installation](#installation)\n  - [Requirements](#requirements)\n- [Development](#development)\n- [Configuration](#configuration)\n- [Support](#support)\n- [Maintainer](#maintainer)\n- [Contributing](#contributing)\n- [License](#license)\n\n## Installation\n\n### Requirements\n\nPython 3.7 and later is supported.\n\nBesides Python Notus Scanner also needs to have\n\n- paho-mqtt\n- psutil\n- python-gnupg\n\ninstalled.\n\n## Development\n\n**notus-scanner** uses [poetry] for its own dependency management and build\nprocess.\n\nFirst install poetry via pip\n\n    python3 -m pip install --user poetry\n\nAfterwards run\n\n    poetry install\n\nin the checkout directory of **notus-scanner** (the directory containing the\n`pyproject.toml` file) to install all dependencies including the packages only\nrequired for development.\n\nAfterwards activate the git hooks for auto-formatting and linting via\n[autohooks].\n\n    poetry run autohooks activate\n\nValidate the activated git hooks by running\n\n    poetry run autohooks check\n\n## Configuration\n\nThe configuration of notus-scanner can be done by providing a TOML config file.\nPer default notus-scanner tries to load the settings from config files in the\nfollowing order: `~/.config/notus-scanner.toml`, `/etc/gvm/notus-scanner.toml`.\n\nAlternatively the location of the to be loaded config file can be set via the\n`-c`/`--config` command line argument. Setting a config file via command line\nwill ignore the default config files.\n\nThe settings are read from a `[notus-scanner]` [section](https://toml.io/en/v1.0.0#table).\n\nExample config file:\n```toml\n[notus-scanner]\nmqtt-broker-address = "1.2.3.4"\nmqtt-broker-port = "1234"\nproducts-directory = "/tmp/notus/advisories/products"\npid-file = "/tmp/notus-scanner.pid"\nlog-file = "/tmp/notus-scanner.log"\nlog-level = "DEBUG"\ndisable-hashsum-verification = true\n```\n\nEach setting can be overridden via an environment variable or command line\nargument.\n\n|Config|Environment|Default|Description|\n|------|-----------|-------|-----------|\n|log-file|NOTUS_SCANNER_LOG_FILE|syslog|File for log output|\n|log-level|NOTUS_SCANNER_LOG_LEVEL|INFO|Minimum level for log output|\n|mqtt-broker-address|NOTUS_SCANNER_MQTT_BROKER_ADDRESS|localhost|IP or DNS address of the MQTT broker|\n|mqtt-broker-port|NOTUS_SCANNER_MQTT_BROKER_PORT|1883|Port of the MQTT broker|\n|pid-file|NOTUS_SCANNER_PID_FILE|/run/notus-scanner/notus-scanner.pid|File for storing the process ID|\n|products-directory|NOTUS_SCANNER_PRODUCTS_DIRECTORY|/var/lib/openvas/plugins/notus/products|Directory for loading product advisories|\n|disable-hashsum-verification| NOTUS_DISABLE_HASHSUM_VERIFICATION | To disable hashsum verification of products |\n\n## Support\n\nFor any question on the usage of Notus Scanner please use the\n[Greenbone Community Portal]. If you found a problem with the software, please\ncreate an issue on GitHub. If you are a Greenbone customer you may alternatively\nor additionally forward your issue to the Greenbone Support Portal.\n\n## Maintainer\n\nThis project is maintained by [Greenbone Networks GmbH][Greenbone Networks]\n\n## Contributing\n\nYour contributions are highly appreciated. Please\n[create a pull request](https://github.com/greenbone/notus-scanner/pulls)\non GitHub. Bigger changes need to be discussed with the development team via the\n[issues section at GitHub](https://github.com/greenbone/notus-scanner/issues)\nfirst.\n\n## License\n\nCopyright (C) 2021-2022 Greenbone Networks GmbH\n\nLicensed under the GNU Affero General Public License v3.0 or later.\n\n[Greenbone Networks]: https://www.greenbone.net/\n[poetry]: https://python-poetry.org/\n[pip]: https://pip.pypa.io/\n[autohooks]: https://github.com/greenbone/autohooks\n[Greenbone Community Portal]: https://community.greenbone.net/\n',
    'author': 'Greenbone Networks GmbH',
    'author_email': 'info@greenbone.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/greenbone/notus-scanner',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
